 /*
  * This file has been derived from the WebSockets++ project at
  * https://github.com/zaphoyd/websocketpp which is licensed under a BSD-license.
  */

// Copyright (c) 2008-2009 Bjoern Hoehrmann <bjoern@hoehrmann.de>
// See http://bjoern.hoehrmann.de/utf-8/decoder/dfa/ for details.

#ifndef UTF8_VALIDATOR_HPP
#define UTF8_VALIDATOR_HPP

#include <stdint.h>

namespace utf8_validator {

static const unsigned int UTF8_ACCEPT = 0;
static const unsigned int UTF8_REJECT = 1;

static const uint8_t utf8d[] = {
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, // 00..1f
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, // 20..3f
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, // 40..5f
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, // 60..7f
  1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9, // 80..9f
  7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7, // a0..bf
  8,8,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2, // c0..df
  0xa,0x3,0x3,0x3,0x3,0x3,0x3,0x3,0x3,0x3,0x3,0x3,0x3,0x4,0x3,0x3, // e0..ef
  0xb,0x6,0x6,0x6,0x5,0x8,0x8,0x8,0x8,0x8,0x8,0x8,0x8,0x8,0x8,0x8, // f0..ff
  0x0,0x1,0x2,0x3,0x5,0x8,0x7,0x1,0x1,0x1,0x4,0x6,0x1,0x1,0x1,0x1, // s0..s0
  1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,1,1,1,1,1,0,1,0,1,1,1,1,1,1, // s1..s2
  1,2,1,1,1,1,1,2,1,2,1,1,1,1,1,1,1,1,1,1,1,1,1,2,1,1,1,1,1,1,1,1, // s3..s4
  1,2,1,1,1,1,1,1,1,2,1,1,1,1,1,1,1,1,1,1,1,1,1,3,1,3,1,1,1,1,1,1, // s5..s6
  1,3,1,1,1,1,1,3,1,3,1,1,1,1,1,1,1,3,1,1,1,1,1,1,1,1,1,1,1,1,1,1, // s7..s8
};

uint32_t inline
decode(uint32_t* state, uint32_t* codep, uint8_t byte) {
    uint32_t type = utf8d[byte];

    *codep = (*state != UTF8_ACCEPT) ?
        (byte & 0x3fu) | (*codep << 6) :
        (0xff >> type) & (byte);

    *state = utf8d[256 + *state*16 + type];
    return *state;
}

/**
 * A validator for UTF-8 strings
 */
class validator {
    public:
        /// Constuctor
        validator() : m_state(UTF8_ACCEPT),m_codepoint(0) {}

        /** 
         * Validate a single UTF-8 character.
         * @param byte The UTF-8 character to validate
         * @return true, if the character is valid.
         */
        bool consume (uint32_t byte) {
            if (utf8_validator::decode(&m_state,&m_codepoint,byte) == UTF8_REJECT) {
                return false;
            }
            return true;
        }

        template <typename iterator_type>
            /** 
             * Validate a range of UTF-8 characters.
             * @param b Iterator, pointing to the start of the sequence,
             * @param e Iterator, pointing to the end of the sequence.
             * @return true, if the supplied sequence is valid.
             */
            bool decode (iterator_type b, iterator_type e) {
                for (iterator_type i = b; i != e; i++) {
                    if (utf8_validator::decode(&m_state,&m_codepoint,*i) == UTF8_REJECT) {
                        return false;
                    }
                }
                return true;
            }

        /**
         * Retrieve validator state.
         * @return true, if the current state is valid. 
         */
        bool complete() {
            return m_state == UTF8_ACCEPT;
        }

        /// Reset the validator state. 
        void reset() {
            m_state = UTF8_ACCEPT;
            m_codepoint = 0;
        }
    private:
        uint32_t    m_state;
        uint32_t    m_codepoint;
};

// convenience function that creates a validator, validates a complete string 
// and returns the result.
// TODO: should this be inline?
inline bool validate(const std::string& s) {
    validator v;
    if (!v.decode(s.begin(),s.end())) {
        return false;
    }
    return v.complete();
}

} // namespace utf8_validator

#endif // UTF8_VALIDATOR_HPP
